/*
 * mgr.h
 */

#ifndef __LIBNFSNAPI_MGR_H__
#define __LIBNFSNAPI_MGR_H__

/**
 * @defgroup Manager
 *
 * Example usage:
 *	\code
 *	libnfsnapi_mgr_t *mgr;
 *		...
 *	mgr = libnfsnapi_mgr_create("testuser", "p3kxmRKf9dk3l6ls");
 *		...
 *	libnfsnapi_mgr_destroy(mgr);
 *	\endcode
 *
 * @{
 */

#include "error.h"

/** Manager handle. Users should treat this structure as opaque. */
typedef struct {
#ifdef LIBNFSNAPI_BUILDING_LIB
	/* Opaque! Don't touch anything here! */
	libnfsnapi_error_t error;
	char *login, *api_key;

	int nudge;			/* time nudge */

	struct CURL *curl;		/* Persistent libcurl handle */

	char *buf;			/* Incoming data buffer */
	size_t buf_read, buf_sz;	/* Data loaded, buffer size */
	struct curl_slist *headers;	/* HTTP headers */
#endif
} libnfsnapi_mgr_t;


#ifdef LIBNFSNAPI_BUILDING_LIB
/* HTTP methods */
#define LIBNFSNAPI_MGR_HTTP_GET		1
#define LIBNFSNAPI_MGR_HTTP_PUT		2
#define LIBNFSNAPI_MGR_HTTP_POST	3
#endif


/**
 * Create a Manager object, used for accessing the NFSN API.
 *
 * This should only ever fail on \c malloc failure. It \em could fail whilst
 * initialising libcurl, but that's pretty unlikely.
 *
 * @param login Your NFSN member name. It will be copied internally.
 * @param api_key Your NFSN-issued API key. It will be copied internally.
 * @param error_out If not NULL, and an error occurs, an error code will be
 *		stored at this memory location (see \ref libnfsnapi_error_t).
 * \return A Manager object, or NULL on failure.
 */
libnfsnapi_mgr_t *libnfsnapi_mgr_create(const char *login, const char *api_key,
					int *error_out);
/**
 * Destroy a Manager object, and free all memory associated with it.
 * Do not use this pointer after calling this function.
 */
void libnfsnapi_mgr_destroy(libnfsnapi_mgr_t *mgr);
/**
 * Return the error code for the last failure associated with this Manager
 * object. This will only be a sensible value if a libNFSNapi call actually
 * indicated failure to its caller somehow.
 *
 * \see Error
 */
int libnfsnapi_mgr_error(libnfsnapi_mgr_t *mgr);
/**
 * Set the time nudge for this Manager object.
 *
 * Interfacing with the NFSN API requires a timestamp to be used, and it must
 * be accurate to within 5 seconds. Ideally, the client of this library would
 * be running on a machine with an NTP-synchronised clock. However, if that is
 * not possible or desired, the time nudge setting can be used to artificially
 * alter the generated timestamp.
 *
 * @param mgr The Manager object.
 * @param nudge Number of seconds to add to each generated timestamp.
 */
void libnfsnapi_mgr_set_timenudge(libnfsnapi_mgr_t *mgr, int nudge);

#ifdef LIBNFSNAPI_BUILDING_LIB	/** \internal */
/* Private functions */
int libnfsnapi_mgr_do(libnfsnapi_mgr_t *mgr, int method,
			const char *request_uri, const char *data);
/* (private) These should only be called after a successful _mgr_do */
long libnfsnapi_mgr_result_long(libnfsnapi_mgr_t *mgr);
float libnfsnapi_mgr_result_float(libnfsnapi_mgr_t *mgr);
char *libnfsnapi_mgr_result_string(libnfsnapi_mgr_t *mgr);
#endif

/** @} */

#endif /* __LIBNFSNAPI_MGR_H__ */
